@@ -1,6 +1,6 @@
 package MangoX::Queue;
 
-use Mojo::Base -base;
+use Mojo::Base 'Mojo::EventEmitter';
 
 use Carp 'croak';
 use DateTime;
@@ -11,7 +11,7 @@ use MangoX::Queue::Delay;
 
 no warnings 'experimental::smartmatch';
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 # A logger
 has 'log' => sub { Mojo::Log->new->level('error') };
@@ -28,6 +28,12 @@ has 'timeout' => sub { $ENV{MANGOX_QUEUE_JOB_TIMEOUT} // 60 };
 # How many times to retry a job before giving up
 has 'retries' => sub { $ENV{MANGOX_QUEUE_JOB_RETRIES} // 5 };
 
+# Store Mojo::IOLoop->timer IDs
+has 'consumers' => sub { {} };
+
+# Store plugins
+has 'plugins' => sub { {} };
+
 sub new {
 	my $self = shift->SUPER::new(@_);
 
@@ -36,6 +42,25 @@ sub new {
 	return $self;
 }
 
+sub plugin {
+	my ($self, $name, $options) = @_;
+
+	croak qq{Plugin $name already loaded} if exists $self->plugins->{$name};
+
+	{
+		no strict 'refs';
+		unless($name->can('new')) {
+			eval "require $name" or croak qq{Failed to load plugin $name: $@};
+		}
+	}
+
+	$self->plugins->{$name} = $name->new(%$options);
+
+	$self->plugins->{$name}->register($self);
+
+	return $self->plugins->{$name};
+}
+
 sub get_options {
 	my ($self) = @_;
 
@@ -93,32 +118,43 @@ sub enqueue {
 		attempt => 1,
 	};
 
-	# TODO allow non-blocking enqueue
-	my $id = $self->collection->insert($db_job);
-
-	return $id;
+	if($callback) {
+		return $self->collection->insert($db_job => sub {
+			my ($collection, $error, $oid) = @_;
+			$db_job->{_id} = $oid;
+			$self->emit_safe(enqueued => $db_job) if $self->has_subscribers('enqueued');
+			$callback->($db_job);
+		});
+	} else {
+		my $id = $self->collection->insert($db_job);
+		$db_job->{_id} = $id;
+		$self->emit_safe(enqueued => $db_job) if $self->has_subscribers('enqueued');
+		return $db_job;
+	}
 }
 
-sub monitor {
-	my ($self, $id, $status, $callback) = @_;
+sub watch {
+	my ($self, $id_or_job, $status, $callback) = @_;
+
+	my $id = ref($id_or_job) ? $id_or_job->{_id} : $id_or_job;
 
 	$status //= 'Complete';
 
 	# args
-	# - wait $queue $id, 'Status' => $callback
+	# - watch $queue $id, 'Status' => $callback
 
 	if($callback) {
 		# Non-blocking
 		$self->log->debug("Waiting for $id on status $status in non-blocking mode");
-		return Mojo::IOLoop->timer(0 => sub { $self->_monitor_nonblocking($id, $status, $callback) });
+		return Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($id, $status, $callback) });
 	} else {
 		# Blocking
 		$self->log->debug("Waiting for $id on status $status in blocking mode");
-		return $self->_monitor_blocking($id, $status);
+		return $self->_watch_blocking($id, $status);
 	}
 }
 
-sub _monitor_blocking {
+sub _watch_blocking {
 	my ($self, $id, $status) = @_;
 
 	while(1) {
@@ -133,7 +169,7 @@ sub _monitor_blocking {
 	}
 }
 
-sub _monitor_nonblocking {
+sub _watch_nonblocking {
 	my ($self, $id, $status, $callback) = @_;
 
 	$self->collection->find_one({'_id' => $id} => sub {
@@ -148,7 +184,7 @@ sub _monitor_nonblocking {
 			$self->log->debug("Job not found or status doesn't match");
 			$self->delay->wait(sub {
 				return unless Mojo::IOLoop->is_running;
-				Mojo::IOLoop->timer(0 => sub { $self->_monitor_nonblocking($id, $status, $callback) });
+				Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($id, $status, $callback) });
 			});
 			return undef;
 		}
@@ -156,69 +192,126 @@ sub _monitor_nonblocking {
 }
 
 sub requeue {
-	my ($self, $job) = @_;
+	my ($self, $job, $callback) = @_;
 
 	$job->{status} = 'Pending';
-	my $id = $self->collection->update({'_id' => $job->{_id}}, $job, {upsert => 1});
-
-	return $id;
-
-	# TODO non-blocking
+	return $self->update($job, $callback);
 }
 
 sub dequeue {
-	my ($self, $id) = @_;
+	my ($self, $id_or_job, $callback) = @_;
 
 	# TODO option to not remove on dequeue?
-	# TODO non-blocking
 
-	$self->collection->remove({'_id' => $id});
+	my $id = ref($id_or_job) ? $id_or_job->{_id} : $id_or_job;
+
+	if($callback) {
+		$self->collection->remove({'_id' => $id} => sub {
+			$callback->();
+			$self->emit_safe(dequeued => $id_or_job) if $self->has_subscribers('dequeued');
+		});
+	} else {
+		$self->collection->remove({'_id' => $id});
+		$self->emit_safe(dequeued => $id_or_job) if $self->has_subscribers('dequeued');
+	}
 }
 
 sub get {
-	my ($self, $id) = @_;
+	my ($self, $id_or_job, $callback) = @_;
 
-	# TODO non-blocking
+	my $id = ref($id_or_job) ? $id_or_job->{_id} : $id_or_job;
 
-	return $self->collection->find_one({'_id' => $id});
+	if($callback) {
+		return $self->collection->find_one({'_id' => $id} => sub {
+			my ($collection, $error, $doc) = @_;
+			$callback->($doc);
+		});
+	} else {
+		return $self->collection->find_one({'_id' => $id});
+	}
+}
+
+sub update {
+	my ($self, $job, $callback) = @_;
+
+	if($callback) {
+		return $self->collection->find_one({'_id' => $job->{_id}} => sub {
+			my ($collection, $error, $doc) = @_;
+			$callback->($doc);
+		});
+	} else {
+		return $self->collection->update({'_id' => $job->{_id}}, $job, {upsert => 1});
+	}
 }
 
 sub fetch {
-	my ($self, $callback) = @_;
+	my ($self, @args) = @_;
+
+	# fetch $queue status => 'Complete', sub { my $job = shift; }
+
+	my $callback = ref($args[-1]) eq 'CODE' ? pop @args : undef;
+	my %args;
+	%args = (@args) if scalar @args;
 
 	$self->log->debug("In fetch");
 
 	if($callback) {
 		$self->log->debug("Fetching in non-blocking mode");
-		return Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($callback, 1) });
+		my $consumer_id = (scalar keys %{$self->consumers}) + 1;
+		$self->consumers->{$consumer_id} = Mojo::IOLoop->timer(0 => sub { $self->_consume_nonblocking(\%args, $consumer_id, $callback, 1) });
+		return $consumer_id;
 	} else {
 		$self->log->debug("Fetching in blocking mode");
-		return $self->_watch_blocking(1);
+		return $self->_consume_blocking(\%args, 1);
 	}
 }
 
-sub watch {
-	my ($self, $callback) = @_;
+sub consume {
+	my ($self, @args) = @_;
 
-	$self->log->debug("In watch");
+	# consume $queue status => 'Failed', sub { my $job = shift; }
+
+	my $callback = ref($args[-1]) eq 'CODE' ? pop @args : undef;
+	my %args;
+	%args = (@args) if scalar @args;
+
+	$self->log->debug("In consume");
 
 	if($callback) {
-		$self->log->debug("Watching in non-blocking mode");
-		return Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($callback, 0) });
+		$self->log->debug("consuming in non-blocking mode");
+		my $consumer_id = (scalar keys %{$self->consumers}) + 1;
+		$self->consumers->{$consumer_id} = Mojo::IOLoop->timer(0 => sub { $self->_consume_nonblocking(\%args, $consumer_id, $callback, 0) });
+		$self->log->debug("Timer scheduled, consumer_id $consumer_id has timer ID: " . $self->consumers->{$consumer_id});
+		return $consumer_id;
 	} else {
-		$self->log->debug("Watching in blocking mode");
-		return $self->_watch_blocking(0);
+		$self->log->debug("consuming in blocking mode");
+		return $self->_consume_blocking(\%args, 0);
 	}
 }
 
-sub _watch_blocking {
-	my ($self, $fetch) = @_;
+sub release {
+	my ($self, $consumer_id) = @_;
+
+	$self->log->debug("Releasing consumer $consumer_id with timer ID: " . $self->consumers->{$consumer_id});
+
+	Mojo::IOLoop->remove($self->consumers->{$consumer_id});
+	delete $self->consumers->{$consumer_id};
+
+	return;
+}
+
+sub _consume_blocking {
+	my ($self, $args, $fetch) = @_;
 
 	while(1) {
-		my $doc = $self->collection->find_and_modify($self->get_options);
+		my $opts = $self->get_options;
+		$opts->{query} = $args if scalar keys %$args;
+
+		my $doc = $self->collection->find_and_modify($opts);
 		$self->log->debug("Job found by Mango: " . ($doc ? 'Yes' : 'No'));
 
 		if($doc) {
+			$self->emit_safe(consumed => $doc) if $self->has_subscribers('consumed');
 			return $doc;
 		} else {
 			last if $fetch;
@@ -227,24 +320,32 @@ sub _watch_blocking {
 	}
 }
 
-sub _watch_nonblocking {
-	my ($self, $callback, $fetch) = @_;
+sub _consume_nonblocking {
+	my ($self, $args, $consumer_id, $callback, $fetch) = @_;
 
-	$self->collection->find_and_modify($self->get_options => sub {
+	my $opts = $self->get_options;
+	$opts->{query} = $args if scalar keys %$args;
+
+	$self->collection->find_and_modify($opts => sub {
 		my ($cursor, $err, $doc) = @_;
 		$self->log->debug("Job found by Mango: " . ($doc ? 'Yes' : 'No'));
 		
 		if($doc) {
 			$self->delay->reset;
+			$self->emit_safe(consumed => $doc) if $self->has_subscribers('consumed');
 			$callback->($doc);
 			return unless Mojo::IOLoop->is_running;
 			return if $fetch;
-			Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($callback) });
+			return unless exists $self->consumers->{$consumer_id};
+			$self->consumers->{$consumer_id} = Mojo::IOLoop->timer(0 => sub { $self->_consume_nonblocking($args, $consumer_id, $callback, 0) });
+			$self->log->debug("Timer rescheduled (recursive immediate), consumer_id $consumer_id has timer ID: " . $self->consumers->{$consumer_id});
 		} else {
+			return unless Mojo::IOLoop->is_running;
+			return if $fetch;
 			$self->delay->wait(sub {
-				return unless Mojo::IOLoop->is_running;
-				return if $fetch;
-				Mojo::IOLoop->timer(0 => sub { $self->_watch_nonblocking($callback) });
+				return unless exists $self->consumers->{$consumer_id};
+				$self->consumers->{$consumer_id} = Mojo::IOLoop->timer(0 => sub { $self->_consume_nonblocking($args, $consumer_id, $callback, 0) });
+				$self->log->debug("Timer rescheduled (recursive delayed), consumer_id $consumer_id has timer ID: " . $self->consumers->{$consumer_id});
 			});
 			return undef;
 		}
@@ -259,6 +360,15 @@ sub _watch_nonblocking {
 
 MangoX::Queue - A MongoDB queue implementation using Mango
 
+=head1 DESCRIPTION
+
+L<MangoX::Queue> is a MongoDB backed queue implementation using L<Mango> to support
+blocking and non-blocking queues.
+
+L<MangoX::Queue> makes no attempt to handle the L<Mango> connection, database or
+collection - pass in a collection to the constructor and L<MangoX::Queue> will
+use it. The collection can be plain, capped or sharded.
+
 =head1 SYNOPSIS
 
 	use Mango;
@@ -269,116 +379,63 @@ MangoX::Queue - A MongoDB queue implementation using Mango
 
 	my $queue = MangoX::Queue->new(collection => $collection);
 
-	# To add a basic job
-	enqueue $queue 'some job name';
-	$queue->enqueue('some job name');
+	# To add a job
+	my $id = enqueue $queue 'test'; # Blocking
+	enqueue $queue 'test' => sub { my $id = shift; }; # Non-blocking
 
-	# To add a complex job
-	enqueue $queue +{
-		foo => 'bar'
-	};
-	$queue->enqueue({
-		foo => 'bar'
-	});
+	# To set options
+	my $id = enqueue $queue priority => 1, created => DateTime->now, 'test'; # Blocking
+	enqueue $queue priority => 1, created => DateTime->now, 'test' => sub { my $id = shift; }; # Non-blocking
 
-	# To set priority
-	enqueue $queue priority => 2, 'job_name';
-	$queue->enqueue(priority => 2, 'job_name');
-
-	# To set created
-	enqueue $queue created => DateTime->now, 'job_name';
-	$queue->enqueue(created => DateTime->now, 'job_name');
-
-	# To set status
-	enqueue $queue status => 'Pending', 'job_name';
-	$queue->enqueue(status => 'Pending', 'job_name');
-
-	# To set multiple options
-	enqueue $queue priority => 1, created => DateTime->now, 'job_name';
-	$queue->enqueue(priority => 1, created => DateTime->now, 'job_name');
-
-	# To wait for a job status change (non-blocking)
-	my $id = enqueue $queue 'test';
-	monitor $queue $id, 'Complete' => sub {
+	# To watch for a specific job status
+	watch $queue $id; # Blocking
+	watch $queue $id, 'Complete' => sub { # Non-blocking
 		# Job status is 'Complete'
 	};
 
-	# To wait on mutliple statuses (non-blocking)
-	my $id = enqueue $queue 'test';
-	monitor $queue $id, ['Complete', 'Failed'] => sub {
-		# Job status is 'Complete' or 'Failed'
-	};
-
-	# To wait for a job status change (blocking)
-	my $id = enqueue $queue 'test';
-	monitor $queue $id, 'Complete';
-
-	# To fetch a job (blocking)
-	my $job = fetch $queue;
-	my $job = $queue->fetch;
-
-	# To fetch a job (non-blocking)
-	fetch $queue sub {
+	# To fetch a job
+	my $job = fetch $queue; # Blocking
+	fetch $queue sub { # Non-blocking
 		my ($job) = @_;
 		# ...
 	};
-	$queue->fetch(sub {
-		my ($job) = @_;
-		# ...
-	});
 
-	# To get a job by id (currently blocking)
-	my $id = enqueue $queue 'test';
-	my $job = get $queue $id;
+	# To get a job by id
+	my $job = get $queue $id; # Blocking
+	get $queue $id => sub { my $job = shift; }; # Non-blocking
 
-	# To requeue a job (currently blocking)
-	my $id = enqueue $queue 'test';
-	my $job = get $queue $id;
-	requeue $queue $job;
+	# To requeue a job
+	my $id = requeue $queue $job; # Blocking
+	requeue $queue $job => sub { my $id = shift; }; # Non-blocking
 
-	# To dequeue a job (currently blocking)
-	my $id = enqueue $queue 'test';
-	dequeue $queue $id;
+	# To dequeue a job
+	dequeue $queue $id; # Blocking
+	dequeue $queue $id => sub { }; # Non-blocking
 
-	# To watch a queue (blocking)
-	while (my $job = watch $queue) {
+	# To consume a queue
+	while(my $job = consume $queue) { # Blocking
 		# ...
 	}
-	while (my $job = $queue->watch) {
-		# ...
-	}
-
-	# To watch a queue (non-blocking)
-	watch $queue sub {
+	my $consumer = consume $queue sub { # Non-blocking
 		my ($job) = @_;
 		# ...
 	};
-	$queue->watch(sub{
-		my ($job) = @_;
-		# ...
-	});
 
-=head1 DESCRIPTION
+	# To stop consuming a queue
+	release $queue $consumer;
 
-L<MangoX::Queue> is a MongoDB backed queue implementation using L<Mango> to support
-blocking and non-blocking queues.
+	# To listen for events
+	on $queue enqueued => sub ( my ($queue, $job) = @_; };
+	on $queue dequeued => sub ( my ($queue, $job) = @_; };
+	on $queue consumed => sub { my ($queue, $job) = @_; };
 
-L<MangoX::Queue> makes no attempt to handle the L<Mango> connection, database or
-collection - pass in a collection to the constructor and L<MangoX::Queue> will
-use it. The collection can be plain, capped or sharded.
+	# To register a plugin
+	plugin $queue 'MangoX::Queue::Plugin::Statsd';
 
 =head1 ATTRIBUTES
 
 L<MangoX::Queue> implements the following attributes.
 
-=head2 delay
-
-	my $delay = $queue->delay;
-	$queue->delay(MangoX::Queue::Delay->new);
-
-The L<MangoX::Queue::Delay> responsible for dynamically controlling the
-delay between queue queries.
-
 =head2 collection
 
     my $collection = $queue->collection;
@@ -388,6 +445,20 @@ delay between queue queries.
 
 The L<Mango::Collection> representing the MongoDB queue collection.
 
+=head2 delay
+
+	my $delay = $queue->delay;
+	$queue->delay(MangoX::Queue::Delay->new);
+
+The L<MangoX::Queue::Delay> responsible for dynamically controlling the
+delay between queue queries.
+
+=head2 plugins
+
+	my $plugins = $queue->plugins;
+
+Returns a hash containing the plugins registered with this queue.
+
 =head2 retries
 
 	my $retries = $queue->retries;
@@ -404,10 +475,65 @@ marked as failed.
 The time (in seconds) a job is allowed to stay in Retrieved state before
 it is released back into Pending state. Defaults to 60 seconds.
 
+=head1 EVENTS
+
+L<MangoX::Queue> inherits from L<Mojo::EventEmitter> and emits the following events
+
+=head2 consumed
+
+	on $queue consumed => sub {
+		my ($queue, $job) = @_;
+		# ...
+	};
+
+Emitted when an item is consumed (either via consume or fetch)
+
+=head2 dequeued
+
+	on $queue dequeued => sub {
+		my ($queue, $job) = @_;
+		# ...
+	};
+
+Emitted when an item is dequeued
+
+=head2 enqueued
+
+	on $queue enqueued => sub {
+		my ($queue, $job) = @_;
+		# ...
+	};
+
+Emitted when an item is enqueued
+
 =head1 METHODS
 
 L<MangoX::Queue> implements the following methods.
 
+=head2 consume
+
+	# In blocking mode
+	while(my $job = consume $queue) {
+		# ...
+	}
+	while(my $job = $queue->consume) {
+		# ...
+	}
+
+	# In non-blocking mode
+	consume $queue sub {
+		my ($job) = @_;
+		# ...
+	};
+	$queue->consume(sub {
+		my ($job) = @_;
+		# ...
+	});
+
+Waits for jobs to arrive on the queue, sleeping between queue checks using L<MangoX::Queue::Delay> or L<Mojo::IOLoop>.
+
+Currently sets the status to 'Retrieved' before returning the job.
+
 =head2 dequeue
 
 	my $job = fetch $queue;
@@ -462,19 +588,16 @@ Gets a job from the queue by ID. Doesn't change the job status.
 Returns the L<Mango::Collection> options hash used by find_and_modify to
 identify and update available queue items.
 
-=head2 monitor
+Wait for a job to enter a certain status.
 
-	# In blocking mode
-	my $id = enqueue $queue 'test';
-	monitor $queue $id, 'Complete'; # blocks until job is complete
+=head2 release
 
-	# In non-blocking mode
-	my $id = enqueue $queue 'test';
-	monitor $queue $id, 'Complete' => sub {
+	my $consumer = consume $queue sub {
 		# ...
 	};
+	release $queue $consumer;
 
-Wait for a job to enter a certain status.
+Releases a non-blocking consumer from watching a queue.
 
 =head2 requeue
 
@@ -483,29 +606,25 @@ Wait for a job to enter a certain status.
 
 Requeues a job. Sets the job status to 'Pending'.
 
+=head2 update
+
+	my $job = fetch $queue;
+	$job->{status} = 'Failed';
+	update $queue $job;
+
+Updates a job in the queue.
+
 =head2 watch
 
 	# In blocking mode
-	while(my $job = watch $queue) {
-		# ...
-	}
-	while(my $job = $queue->watch) {
-		# ...
-	}
+	my $id = enqueue $queue 'test';
+	watch $queue $id, 'Complete'; # blocks until job is complete
 
 	# In non-blocking mode
-	watch $queue sub {
-		my ($job) = @_;
+	my $id = enqueue $queue 'test';
+	watch $queue $id, 'Complete' => sub {
 		# ...
 	};
-	$queue->watch(sub {
-		my ($job) = @_;
-		# ...
-	});
-
-Watches the queue for jobs, sleeping between queue checks using L<MangoX::Queue::Delay>.
-
-Currently sets job status to 'Retrieved'.
 
 =head1 SEE ALSO
 
